/*  2016 Script by Taylor Mouse
    Game: Strife
    Url: https://strife.com/
    Release notes:
    0.1 - Initial setup
        - Imports vertices, tris, uv's, normals, bones, vertex weights
        - Layers to clean up 
        
    TODO: 
        - import material from external file
        - create utility for this
        - import animations
*/
ClearListener()
/** Helper functions ***************************************/
function ReadFixedString bin nbr = 
(
    local str = ""
    
    for i=1 to nbr do
    (
        s= bit.IntAsChar(ReadByte bin)

        if ( s !="\0") then str+=s
    )
    return str
)
function SkipBytes bin nBytesToSkip=
(
	/* Skip n-bytes without reading them into memory */
	fseek bin nBytesToSkip #seek_cur
)
fn GetLayer layerName =
(
	/* Create layer in MAX, if it exists, just return a reference to that layer */
	layer = LayerManager.GetLayerFromName layerName
	if(layer == undefined) then
	(
		layer = LayerManager.NewLayerFromName layerName
	)
	return layer
)
/** Main Read piece ****************************************/
function ReadStrifeModel fileName =
(
    if( DoesFileExist(fileName) == false) then throw ( "File not found Exception: " + fileName )
    
    local fileSize = getFileSize fileName
    local uniqueChunks = #()
    
    /* STRUCTURES */    
    struct StrifeChunk 
    ( 
        /* chunks found in the files:
            head
            bone
            mesh
            vrts
            lnk3 -> holds the vertex/bone weights
            face
            texc
            tang -> not known to be used in max
            sign -> UNKNOWN ( for now )
            nrml
        */
        Name, Offset, NbrItems, Size,
        
        function Read bin theName =
        (
            Name = theName
            Size = ReadLong bin
            OffSet = fTell bin
            fSeek bin Size #seek_cur
        )
    )
    
    struct StrifeBone ( Name, ParentBoneId, theBone, Position )
    
    struct VertexWeight ( BoneIds, Weights )
    
    /* Open file to start reading its data */
    local stream = fOpen fileName "rb"
    
    /* Magic SMDL*/
    local smdl = (ReadFixedString stream 4)
    
    if(smdl !="SMDL" ) then throw "File is not the correct file type!"
    
    /* Read in all the chunks */
    local allChunks = #()
    
    while (fileSize > (fTell stream) ) do
    (
        local chunkName = (ReadFixedString stream 4)
        local chnk = StrifeChunk()
        
        chnk.Read stream chunkName
        
        append allChunks chnk
        appendIfUnique uniqueChunks chunkName
    )
    
    /* Process all chunks */
    local verts = #()
    local tris = #()
    local uvs = #()
    local norms = #()
    local allBones = #()
    local vertexWeights = #() --> weight of bone that influences the per vertex 
    local weightedBones = #() --> bones that will be added to the skin of the model
    
    
    local meshName = ""
    local materialName = ""
    local nbrBones = 0
    local nbrMeshes = 0
    
    local meshLayer = GetLayer "Mesh"
    local boneLayer = GetLayer "Bones"
    
    for chunk in allChunks do
    (
        if( chunk.Name == "head" ) then
        (
            fSeek stream chunk.Offset #seek_set
            
            readLong stream
            nbrMeshes = readLong stream
            readLong stream
            readLong stream
            
            nbrBones = readLong stream
            
            format "Number of meshes : %\n" nbrMeshes
            format "Number of bones  : %\n" nbrBones
            
            /* bounding box */
            for i=1 to 6 do readFloat stream
            
        )
                
        /* read the mesh */
        if( chunk.Name =="mesh") then
        (
            fSeek stream chunk.Offset #seek_set
            SkipBytes stream 36 /* presumably a 3x3 float matrix */
            SkipBytes stream 4 /* FF FF FF FF */
            local nameLen = readByte stream
            local materialLen = readByte stream
            meshName = ReadFixedString stream nameLen
            readByte stream
            materialName = ReadFixedString stream materialLen
            
        )
        
        if( chunk.Name == "bone") then
        (
            fSeek stream chunk.Offset #seek_set
            SkipBytes stream 4 /* FF FF FF FF */
            SkipBytes stream 96 /* presumably a 6x4 matrix */
            
            
            
            /* read and build the bones */
            for b=1 to nbrBones do
            (
                len = readByte stream
                
                b = StrifeBone()
                
                b.Name = readFixedString stream len
                readByte stream
                
                b.ParentBoneId = readLong stream 
                
                SkipBytes stream 36
                SkipBytes stream 48
                x = readfloat stream 
                y = readfloat stream
                z = readfloat stream  
                
                local theBone = BoneSys.CreateBone [0,0,0] [0,0,0] [0,0,1]
                theBone.ShowLinks = true
                theBone.Width = 0.05
                theBone.Height = 0.05
                theBone.Name = b.Name
                theBone.Position = [x,y,z]
                b.theBone = theBone
                append allBones b
                
                boneLayer.AddNode theBone
                
            )
            /* create the bone hierarchy */
            for b = 1 to nbrBones-1 do
            (
                if(allBones[b].ParentBoneId != 0) then
                (
                    local parentBone = allBones[ allBones[b].ParentBoneId ].TheBone
                    if(parentBone != undefined and allBones[b].TheBone != undefined ) then
                    (
                        allBones[b].TheBone.Parent = parentBone
                    )
                )
                
            )
            
            
        )
        
        /* Read the vertices */
        if( chunk.Name == "vrts") then
        (
            verts = #()
            
            fSeek stream chunk.Offset #seek_set
            SkipBytes stream 4
            chunk.NbrItems = (chunk.Size - 4 ) / 12
            
            for i=1 to chunk.NbrItems do
            (
                append verts [ readFloat stream, readFloat stream, readFloat stream ]
            )
        )

        /* Read the lnk3 data ?? */
        if( chunk.Name == "lnk3") then
        (
            vertexWeights = #()
            weightedBones = #()
            
            fSeek stream chunk.Offset #seek_set
            
            SkipBytes stream 4
            
            chunk.NbrItems = ReadLong stream --> matches the nbr of vertices
            
           
            for l=1 to chunk.NbrItems do
            (
                local vw = VertexWeight() 
                local boneIds = #()
                local weights = #()
                local nbrWeights = ReadLong stream
                
                if (nbrWeights > 4 ) then throw "Please check, a vertex should not have more than 4 weights !"

                for w=1 to nbrWeights do
                (
                    append weights (readfloat stream)
                )
                
                for w=1 to nbrWeights do
                (
                    local bid = (ReadLong stream ) + 1
                    
                    append boneIds bid
                    
                    appendIfUnique weightedBones bid
                )
                
                vw.BoneIds = boneIds
                vw.Weights = weights
            
                append vertexWeights vw
                
            )

        )
        
        /* Read the sign data ?? */
        if( chunk.Name == "sign") then
        (
            --fSeek stream chunk.Offset #seek_set
            
            --format "Size of sign chunk : %\n" chunk.Size
            
        )
        
        /* Read the faces */
        if( chunk.Name == "face") then
        (
            tris = #()
            
            fSeek stream chunk.Offset #seek_set
            local unk = readLong stream
            
            chunk.NbrItems = readLong stream

            format "Number of faces : %\n" chunk.NbrItems
            
            local sizePerIndex = readByte stream 
            
            for f=1 to chunk.NbrItems do
            (
                if(sizePerIndex == 2) then
                (
                    
                    append tris [ readShort stream + 1, readShort stream + 1, readShort stream  + 1]
                )
                
                if(sizePerIndex == 1) then
                (
                    a = readByte stream #unsigned + 1
                    b = readByte stream #unsigned + 1
                    c = readByte stream #unsigned + 1
                    
                    if( a<0 or b<0 or c<0) then throw (" Tris cannot be negative indexed ")
                    
                    
                    
                    append tris [ a, b, c]
                )                
            )

        )
        
        /* read the texture coordinates */
        if (chunk.Name =="texc") then
        (
            uvs = #()
            
            fSeek stream chunk.Offset #seek_set
            SkipBytes stream  8
            chunk.NbrItems = verts.count
            for u=1 to chunk.NbrItems do
            (
                append uvs [readfloat stream, 1.0 - readfloat stream, 1.0]
                
            )
            
            
        )
        
        /* read the normals
            since this is ( currently ) the last chunk in each mesh, build the mesh too 
        */
        if( chunk.Name == "nrml") then
        (
            
            norms = #()
            
            fSeek stream chunk.Offset #seek_set
            SkipBytes stream 4
            chunk.NbrItems = verts.count
            
            for u=1 to chunk.NbrItems do
            (
                append norms [readfloat stream, readfloat stream, readfloat stream]
            )
            
            /* Create the mesh */
            theMesh = mesh name:meshName vertices:verts faces:tris tverts:uvs vnorms:norms
            theMesh.WireColor = Color (random 0 255) (random 0 128) (random 0 255)
            
            /* Apply texture coordinates correctly */
            buildTVFaces theMesh false
            for i = 1 to tris.count do ( setTVFace theMesh i tris[i] )
            update theMesh
            
            /* Apply skin modifier */
            local mshSkin = skin name:"Skin"
            max modify mode
            addmodifier theMesh mshSkin
            modPanel.setCurrentObject mshSkin
            
            /* Add the bones to the model */
            for i=1 to allBones.count do
            (
                local b = allBones[i].theBone
                skinOps.addBone mshSkin b 0
            )   

            update theMesh
            max views redraw
            
            /* Apply the correct weights */
            for v=1 to vertexWeights.count do
            (
                skinOps.ReplaceVertexWeights mshSkin v vertexWeights[v].BoneIds  vertexWeights[v].Weights
            )
            update theMesh
            
            meshLayer.AddNode theMesh
            
        )
        
    )
    
    print uniqueChunks
    
    fClose stream
    return 0
)
/***********************************************************/

fileName = GetOpenFileName caption:"Open a Strife Model file" \
                           types:"Strife model(*.model)|*.model" \
                           historyCategory:"StrifeModelsCategory"

setWaitCursor()
ReadStrifeModel fileName
setArrowCursor()
print "Done!"